/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import type { ICodecOptions } from "../../codec/index.js";
import type { RevisionTag, RevisionTagCodec } from "../rebase/index.js";
import type { FieldKey } from "../schema-stored/index.js";
import type { ProtoNodes, Root } from "./delta.js";
import { DetachedFieldIndex } from "./detachedFieldIndex.js";
import type { NodeIndex, PlaceIndex, Range } from "./pathTree.js";
import { type DeltaVisitor } from "./visitDelta.js";
import type { IIdCompressor } from "@fluidframework/id-compressor";
export declare function makeDetachedFieldIndex(prefix: string | undefined, revisionTagCodec: RevisionTagCodec, idCompressor: IIdCompressor, options?: ICodecOptions): DetachedFieldIndex;
export declare function applyDelta(delta: Root, latestRevision: RevisionTag | undefined, deltaProcessor: {
    acquireVisitor: () => DeltaVisitor;
}, detachedFieldIndex: DetachedFieldIndex): void;
export declare function announceDelta(delta: Root, latestRevision: RevisionTag | undefined, deltaProcessor: {
    acquireVisitor: () => DeltaVisitor & AnnouncedVisitor;
}, detachedFieldIndex: DetachedFieldIndex): void;
/**
 * @param visitors - The returned visitor invokes the corresponding events for all these visitors, in order.
 * @param announcedVisitors - Subset of `visitors` to also call {@link AnnouncedVisitor} methods on.
 * This must be a subset of `visitors`: if not the visitor will not have its path correctly set when the events are triggered.
 * When `visitors` are making changes to data, `announcedVisitors` can be used to get extra events before or after all the changes from all the visitors have been made.
 * This can, for example, enable visitors to have access to the tree in these extra events despite multiple separate visitors updating different tree related data-structures.
 * @returns a DeltaVisitor combining all `visitors`.
 */
export declare function combineVisitors(visitors: readonly DeltaVisitor[], announcedVisitors?: readonly AnnouncedVisitor[]): DeltaVisitor;
/**
 * Visitor that is notified of changes before, after, and when changes are made.
 * Must be freed after use.
 */
export interface AnnouncedVisitor extends DeltaVisitor {
    /**
     * A hook that is called after all nodes have been created.
     */
    afterCreate(content: ProtoNodes, destination: FieldKey): void;
    beforeDestroy(field: FieldKey, count: number): void;
    beforeAttach(source: FieldKey, count: number, destination: PlaceIndex): void;
    afterAttach(source: FieldKey, destination: Range): void;
    beforeDetach(source: Range, destination: FieldKey): void;
    afterDetach(source: PlaceIndex, count: number, destination: FieldKey): void;
    beforeReplace(newContent: FieldKey, oldContent: Range, oldContentDestination: FieldKey): void;
    afterReplace(newContentSource: FieldKey, newContent: Range, oldContent: FieldKey): void;
}
/**
 * Creates an announced visitor with only the provided functions and uses a no op for the rest.
 * This is provided to make some of the delta visitor definitions cleaner.
 */
export declare function createAnnouncedVisitor(visitorFunctions: {
    free?: () => void;
    create?: (content: ProtoNodes, destination: FieldKey) => void;
    afterCreate?: (content: ProtoNodes, destination: FieldKey) => void;
    beforeDestroy?: (field: FieldKey, count: number) => void;
    destroy?: (detachedField: FieldKey, count: number) => void;
    beforeAttach?: (source: FieldKey, count: number, destination: PlaceIndex) => void;
    attach?: (source: FieldKey, count: number, destination: PlaceIndex) => void;
    afterAttach?: (source: FieldKey, destination: Range) => void;
    beforeDetach?: (source: Range, destination: FieldKey) => void;
    afterDetach?: (source: PlaceIndex, count: number, destination: FieldKey) => void;
    detach?: (source: Range, destination: FieldKey) => void;
    beforeReplace?: (newContent: FieldKey, oldContent: Range, oldContentDestination: FieldKey) => void;
    replace?: (newContentSource: FieldKey, range: Range, oldContentDestination: FieldKey) => void;
    afterReplace?: (newContentSource: FieldKey, newContent: Range, oldContent: FieldKey) => void;
    enterNode?: (index: NodeIndex) => void;
    exitNode?: (index: NodeIndex) => void;
    enterField?: (key: FieldKey) => void;
    exitField?: (key: FieldKey) => void;
}): AnnouncedVisitor;
//# sourceMappingURL=visitorUtils.d.ts.map