"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.UnsafeUnknownSchema = exports.areFieldSchemaEqual = exports.areImplicitFieldSchemaEqual = exports.normalizeAllowedTypes = exports.normalizeFieldSchema = exports.FieldSchema = exports.createFieldSchema = exports.getDefaultProvider = exports.extractFieldProvider = exports.isConstant = exports.getExplicitStoredKey = exports.getStoredKey = exports.FieldKind = exports.isTreeNodeSchemaClass = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../util/index.js");
const flexList_js_1 = require("./flexList.js");
/**
 * Returns true if the given schema is a {@link TreeNodeSchemaClass}, or otherwise false if it is a {@link TreeNodeSchemaNonClass}.
 * @internal
 */
function isTreeNodeSchemaClass(schema) {
    return schema.constructor !== undefined;
}
exports.isTreeNodeSchemaClass = isTreeNodeSchemaClass;
/**
 * Kind of a field on a node.
 * @remarks
 * More kinds may be added over time, so do not assume this is an exhaustive set.
 * @public
 */
var FieldKind;
(function (FieldKind) {
    /**
     * A field which can be empty or filled.
     * @remarks
     * Allows 0 or one child.
     */
    FieldKind[FieldKind["Optional"] = 0] = "Optional";
    /**
     * A field which must always be filled.
     * @remarks
     * Only allows exactly one child.
     */
    FieldKind[FieldKind["Required"] = 1] = "Required";
    /**
     * A special field used for node identifiers.
     * @remarks
     * Only allows exactly one child.
     */
    FieldKind[FieldKind["Identifier"] = 2] = "Identifier";
})(FieldKind || (exports.FieldKind = FieldKind = {}));
/**
 * Maps from a property key to its corresponding {@link FieldProps.key | stored key} for the provided
 * {@link ImplicitFieldSchema | field schema}.
 *
 * @remarks
 * If an explicit stored key was specified in the schema, it will be used.
 * Otherwise, the stored key is the same as the property key.
 */
function getStoredKey(propertyKey, fieldSchema) {
    return (0, index_js_1.brand)(getExplicitStoredKey(fieldSchema) ?? propertyKey);
}
exports.getStoredKey = getStoredKey;
/**
 * Gets the {@link FieldProps.key | stored key} specified by the schema, if one was explicitly specified.
 * Otherwise, returns undefined.
 */
function getExplicitStoredKey(fieldSchema) {
    return fieldSchema instanceof FieldSchema ? fieldSchema.props?.key : undefined;
}
exports.getExplicitStoredKey = getExplicitStoredKey;
/**
 * Returns true if the given {@link FieldProvider} is a {@link ConstantFieldProvider}
 */
function isConstant(fieldProvider) {
    return fieldProvider.length === 0;
}
exports.isConstant = isConstant;
function extractFieldProvider(input) {
    return input;
}
exports.extractFieldProvider = extractFieldProvider;
function getDefaultProvider(input) {
    return input;
}
exports.getDefaultProvider = getDefaultProvider;
/**
 * All policy for a specific field,
 * including functionality that does not have to be kept consistent across versions or deterministic.
 *
 * This can include policy for how to use this schema for "view" purposes, and well as how to expose editing APIs.
 * Use {@link SchemaFactory} to create the FieldSchema instances, for example {@link SchemaFactory.optional}.
 * @privateRemarks
 * Public access to the constructor is removed to prevent creating expressible but unsupported (or not stable) configurations.
 * {@link createFieldSchema} can be used internally to create instances.
 *
 * @typeParam TCustomMetadata - Custom metadata properties to associate with the field.
 * See {@link FieldSchemaMetadata.custom}.
 *
 * @sealed @public
 */
class FieldSchema {
    /**
     * What types of tree nodes are allowed in this field.
     * @remarks Counterpart to {@link FieldSchema.allowedTypes}, with any lazy definitions evaluated.
     */
    get allowedTypeSet() {
        return this.lazyTypes.value;
    }
    /**
     * {@inheritDoc FieldProps.metadata}
     */
    get metadata() {
        return this.props?.metadata;
    }
    constructor(
    /**
     * The {@link https://en.wikipedia.org/wiki/Kind_(type_theory) | kind } of this field.
     * Determines the multiplicity, viewing and editing APIs as well as the merge resolution policy.
     */
    kind, 
    /**
     * What types of tree nodes are allowed in this field.
     */
    allowedTypes, 
    /**
     * Optional properties associated with the field.
     */
    props) {
        this.kind = kind;
        this.allowedTypes = allowedTypes;
        this.props = props;
        this.lazyTypes = new internal_1.Lazy(() => normalizeAllowedTypes(this.allowedTypes));
        // TODO: optional fields should (by default) get a default provider that returns undefined, removing the need to special case them here:
        this.requiresValue =
            this.props?.defaultProvider === undefined && this.kind !== FieldKind.Optional;
    }
}
exports.FieldSchema = FieldSchema;
(() => {
    exports.createFieldSchema = (kind, allowedTypes, props) => new FieldSchema(kind, allowedTypes, props);
})();
/**
 * Normalizes a {@link ImplicitFieldSchema} to a {@link FieldSchema}.
 */
function normalizeFieldSchema(schema) {
    return schema instanceof FieldSchema
        ? schema
        : (0, exports.createFieldSchema)(FieldKind.Required, schema);
}
exports.normalizeFieldSchema = normalizeFieldSchema;
/**
 * Normalizes a {@link ImplicitAllowedTypes} to a set of {@link TreeNodeSchema}s, by eagerly evaluating any
 * lazy schema declarations.
 *
 * @remarks Note: this must only be called after all required schemas have been declared, otherwise evaluation of
 * recursive schemas may fail.
 *
 * @internal
 */
function normalizeAllowedTypes(types) {
    const normalized = new Set();
    if ((0, index_js_1.isReadonlyArray)(types)) {
        for (const lazyType of types) {
            normalized.add(evaluateLazySchema(lazyType));
        }
    }
    else {
        normalized.add(evaluateLazySchema(types));
    }
    return normalized;
}
exports.normalizeAllowedTypes = normalizeAllowedTypes;
/**
 * Returns true if the given {@link ImplicitFieldSchema} are equivalent, otherwise false.
 * @remarks Two ImplicitFieldSchema are considered equivalent if all of the following are true:
 * 1. They have the same {@link FieldKind | kinds}.
 * 2. They have {@link areFieldPropsEqual | equivalent FieldProps}.
 * 3. They have the same exact set of allowed types. The allowed types must be (respectively) reference equal.
 *
 * For example, comparing an ImplicitFieldSchema that is a {@link TreeNodeSchema} to an ImplicitFieldSchema that is a {@link FieldSchema}
 * will return true if they are the same kind, the FieldSchema has exactly one allowed type (the TreeNodeSchema), and they have equivalent FieldProps.
 */
function areImplicitFieldSchemaEqual(a, b) {
    return areFieldSchemaEqual(normalizeFieldSchema(a), normalizeFieldSchema(b));
}
exports.areImplicitFieldSchemaEqual = areImplicitFieldSchemaEqual;
/**
 * Returns true if the given {@link FieldSchema} are equivalent, otherwise false.
 * @remarks Two FieldSchema are considered equivalent if all of the following are true:
 * 1. They have the same {@link FieldKind | kinds}.
 * 2. They have {@link areFieldPropsEqual | equivalent FieldProps}.
 * 3. They have the same exact set of allowed types. The allowed types must be reference equal.
 */
function areFieldSchemaEqual(a, b) {
    if (a === b) {
        return true;
    }
    if (a.kind !== b.kind) {
        return false;
    }
    if (!areFieldPropsEqual(a.props, b.props)) {
        return false;
    }
    return (0, index_js_1.compareSets)({ a: a.allowedTypeSet, b: b.allowedTypeSet });
}
exports.areFieldSchemaEqual = areFieldSchemaEqual;
/**
 * Returns true if the given {@link FieldProps} are equivalent, otherwise false.
 * @remarks FieldProps are considered equivalent if their keys and default providers are reference equal, and their metadata are {@link areMetadataEqual | equivalent}.
 */
function areFieldPropsEqual(a, b) {
    if (a === b) {
        return true;
    }
    if (a?.key !== b?.key || a?.defaultProvider !== b?.defaultProvider) {
        return false;
    }
    if (!areMetadataEqual(a?.metadata, b?.metadata)) {
        return false;
    }
    return true;
}
/**
 * Returns true if the given {@link FieldSchemaMetadata} are equivalent, otherwise false.
 * @remarks FieldSchemaMetadata are considered equivalent if their custom data and descriptions are (respectively) reference equal.
 */
function areMetadataEqual(a, b) {
    if (a === b) {
        return true;
    }
    return a?.custom === b?.custom && a?.description === b?.description;
}
function evaluateLazySchema(value) {
    const evaluatedSchema = (0, flexList_js_1.isLazy)(value) ? value() : value;
    if (evaluatedSchema === undefined) {
        throw new internal_2.UsageError(`Encountered an undefined schema. This could indicate that some referenced schema has not yet been instantiated.`);
    }
    return evaluatedSchema;
}
/**
 * {@inheritdoc (UnsafeUnknownSchema:type)}
 * @alpha
 */
exports.UnsafeUnknownSchema = Symbol("UnsafeUnknownSchema");
//# sourceMappingURL=schemaTypes.js.map