/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { AllowedUpdateType, type ITreeCursorSynchronous, type TreeStoredSchema } from "../core/index.js";
import type { ITreeCheckout } from "./treeCheckout.js";
import { type ViewSchema } from "../simple-tree/index.js";
/**
 * Modify `storedSchema` and invoke `setInitialTree` when it's time to set the tree content.
 *
 * Requires `storedSchema` to be in its default/empty state.
 *
 * This is done in such a way that if the content (implicitly assumed to start empty)
 * is never out of schema.
 * This means that if the root field of the new schema requires content (like a value field),
 * a temporary intermediate schema is used so the initial empty state is not out of schema.
 *
 * Since this makes multiple changes, callers may want to wrap it in a transaction.
 */
export declare function initializeContent(schemaRepository: {
    storedSchema: ITreeCheckout["storedSchema"];
    updateSchema: ITreeCheckout["updateSchema"];
}, newSchema: TreeStoredSchema, setInitialTree: () => void): void;
export declare enum UpdateType {
    /**
     * Already compatible, no update needed.
     */
    None = 0,
    /**
     * Empty: needs initializing.
     */
    Initialize = 1,
    /**
     * Schema can be upgraded leaving tree as is.
     */
    SchemaCompatible = 2,
    /**
     * No update currently supported.
     */
    Incompatible = 3
}
export declare function evaluateUpdate(viewSchema: ViewSchema, allowedSchemaModifications: AllowedUpdateType, checkout: ITreeCheckout): UpdateType;
export declare function canInitialize(checkout: ITreeCheckout): boolean;
/**
 * Initialize a checkout with a schema and tree content.
 * This function should only be called when the tree is uninitialized (no schema or content).
 * @remarks
 *
 * If the proposed schema (from `treeContent`) is not compatible with the emptry tree, this function handles using an intermediate schema
 * which supports the empty tree as well as the final tree content.
 */
export declare function initialize(checkout: ITreeCheckout, treeContent: TreeStoredContent): void;
/**
 * Ensure a {@link ITreeCheckout} can be used with a given {@link ViewSchema}.
 *
 * @remarks
 * It is up to the caller to ensure that compatibility is reevaluated if the checkout's stored schema is edited in the future.
 *
 * @param viewSchema - View schema that `checkout` should be made compatible with.
 * @param allowedSchemaModifications - Flags enum describing the ways this is allowed to modify `checkout`.
 * @param checkout - To be modified as needed to be compatible with `viewSchema`.
 * @param treeContent - Content to be used to initialize `checkout`'s the tree if needed and allowed.
 * @returns true iff checkout now is compatible with `viewSchema`.
 */
export declare function ensureSchema(viewSchema: ViewSchema, allowedSchemaModifications: AllowedUpdateType, checkout: ITreeCheckout, treeContent: TreeStoredContent | undefined): boolean;
/**
 * Content that can populate a `SharedTree`.
 */
export interface TreeStoredContent {
    readonly schema: TreeStoredSchema;
    /**
     * Default tree content to initialize the tree with iff the tree is uninitialized
     * (meaning it does not even have any schema set at all).
     */
    readonly initialTree: readonly ITreeCursorSynchronous[] | ITreeCursorSynchronous | undefined;
}
//# sourceMappingURL=schematizeTree.d.ts.map