/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { oob } from "@fluidframework/core-utils/internal";
/**
 * If (key1, key2) already has a value in the map, it is returned, otherwise value is added under (key1, key2) and undefined is returned.
 */
export function tryAddToNestedMap(map, key1, key2, value) {
    let innerMap = map.get(key1);
    if (innerMap === undefined) {
        innerMap = new Map();
        map.set(key1, innerMap);
    }
    if (innerMap.has(key2)) {
        return innerMap.get(key2);
    }
    innerMap.set(key2, value);
    return undefined;
}
/**
 * Copies over all entries from the source map into the destination map.
 *
 * @param source - The map to copy data from. Not mutated.
 * @param destination - The map to copy data into. Both the outer and inner map may be mutated.
 * @param override - Whether existing entries in `destination` should be replaced by corresponding entries in `source`.
 *
 * @remarks - This function performs deep copying when necessary.
 * This ensures that mutating `destination` after this call will not result in unexpected mutations to `source`.
 */
export function populateNestedMap(source, destination, override) {
    for (const [key1, sourceInner] of source) {
        let destinationInner = destination.get(key1);
        if (destinationInner === undefined) {
            destinationInner = new Map(sourceInner);
            destination.set(key1, destinationInner);
        }
        else {
            for (const [key2, value] of sourceInner) {
                if (override || !destinationInner.has(key2)) {
                    destinationInner.set(key2, value);
                }
            }
        }
    }
}
/**
 * Sets the value at (key1, key2) in map to value.
 * If there already is a value for (key1, key2), it is replaced with the provided one.
 */
export function setInNestedMap(map, key1, key2, value) {
    const innerMap = getOrAddInMap(map, key1, new Map());
    innerMap.set(key2, value);
}
/**
 * Sets the value at `key` in map to value if not already present.
 * Returns the value at `key` after setting it.
 * This is equivalent to a get or default that adds the default to the map.
 */
export function getOrAddInMap(map, key, value) {
    const currentValue = map.get(key);
    if (currentValue !== undefined) {
        return currentValue;
    }
    map.set(key, value);
    return value;
}
/**
 * Sets the value at `key` in `map` to `generateValue()` if not already present.
 * Returns the value at `key` after setting it.
 */
export function getOrAddInMapLazy(map, key, generateValue) {
    const currentValue = map.get(key);
    if (currentValue !== undefined) {
        return currentValue;
    }
    const value = generateValue();
    map.set(key, value);
    return value;
}
/**
 * Returns the value at (key1, key2) in map, or undefined if not present.
 */
export function tryGetFromNestedMap(map, key1, key2) {
    const innerMap = map.get(key1);
    if (innerMap === undefined) {
        return undefined;
    }
    return innerMap.get(key2);
}
/**
 * If (key1, key2) is not in the map, add value to the map.
 * Returns whatever is at (key1, key2) in map (which will be value if it was empty before).
 */
export function getOrAddInNestedMap(map, key1, key2, value) {
    const existing = tryAddToNestedMap(map, key1, key2, value);
    if (existing !== undefined) {
        return existing;
    }
    return value;
}
/**
 * Does not change map.
 * If (key1, key2) is not in map, returns value.
 * If (key1, key2) is in map, return its entry.
 */
export function getOrDefaultInNestedMap(map, key1, key2, value) {
    const existing = tryGetFromNestedMap(map, key1, key2);
    if (existing !== undefined) {
        return existing;
    }
    return value;
}
/**
 * Removes the value at (key1, key2) from the map.
 *
 * @returns true iff found.
 */
export function deleteFromNestedMap(map, key1, key2) {
    const innerMap = map.get(key1);
    if (innerMap === undefined) {
        return false;
    }
    const deleted = innerMap.delete(key2);
    if (innerMap.size === 0) {
        map.delete(key1);
    }
    return deleted;
}
/**
 * Converts a nested map to a flat list of triplets.
 */
export function nestedMapToFlatList(map) {
    const list = [];
    map.forEach((innerMap, key1) => {
        innerMap.forEach((val, key2) => {
            list.push([key1, key2, val]);
        });
    });
    return list;
}
/**
 * Builds a nested map from a flat list of triplets.
 */
export function nestedMapFromFlatList(list) {
    const map = new Map();
    for (const [key1, key2, val] of list) {
        getOrAddInMap(map, key1, new Map()).set(key2, val);
    }
    return map;
}
export function forEachInNestedMap(map, delegate) {
    map.forEach((innerMap, keyFirst) => {
        innerMap.forEach((val, keySecond) => {
            delegate(val, keyFirst, keySecond);
        });
    });
}
/**
 * Maps the `input` map values using the provided `delegate`.
 *
 * @param input - The `NestedMap` whose contents are being mapped.
 * @param delegate - The delegate to use for mapping values,
 * @returns A new `NestedMap` with the same keys as `input`, but with the values produced by `delegate`.
 */
export function mapNestedMap(input, delegate) {
    const output = new Map();
    input.forEach((inputInnerMap, keyFirst) => {
        const outputInnerMap = new Map();
        inputInnerMap.forEach((val, keySecond) => {
            const mappedValue = delegate(val, keyFirst, keySecond);
            outputInnerMap.set(keySecond, mappedValue);
        });
        output.set(keyFirst, outputInnerMap);
    });
    return output;
}
/**
 * Map with two keys; same semantics as NestedMap, but maintains a size count for the entire collection.
 * Note: undefined is not supported as a value, and will cause incorrect behavior.
 */
export class SizedNestedMap {
    constructor() {
        this.nestedMap = new Map();
        this.count = 0;
    }
    /**
     * Returns the total number of elements in this nested map.
     */
    get size() {
        return this.count;
    }
    /**
     * If (key1, key2) already has a value in the map, it is returned, otherwise value is added under (key1, key2) and undefined is
     * returned.
     */
    tryGet(key1, key2) {
        return tryGetFromNestedMap(this.nestedMap, key1, key2);
    }
    /**
     * Does not change map.
     * If (key1, key2) is not in map, returns value.
     * If (key1, key2) is in map, return its entry.
     */
    getOrDefault(key1, key2, value) {
        return getOrDefaultInNestedMap(this.nestedMap, key1, key2, value);
    }
    /**
     * If (key1, key2) already has a value in the map, it is returned, otherwise value is added under (key1, key2) and undefined is
     * returned.
     */
    tryAdd(key1, key2, value) {
        const currentVal = tryAddToNestedMap(this.nestedMap, key1, key2, value);
        if (currentVal === undefined) {
            this.count++;
        }
        return currentVal;
    }
    /**
     * Sets the value at (key1, key2) in map to value.
     * If there already is a value for (key1, key2), it is replaced with the provided one.
     */
    set(key1, key2, value) {
        if (this.tryAdd(key1, key2, value) !== undefined) {
            setInNestedMap(this.nestedMap, key1, key2, value);
        }
    }
    /**
     * Removes the value at (key1, key2) from the map.
     * Returns true iff found.
     */
    delete(key1, key2) {
        const deleted = deleteFromNestedMap(this.nestedMap, key1, key2);
        if (deleted) {
            this.count--;
        }
        return deleted;
    }
    /**
     * Runs the supplied delegate for every (value, key1, key2).
     */
    forEach(delegate) {
        forEachInNestedMap(this.nestedMap, delegate);
    }
    /**
     * Clears the map.
     */
    clear() {
        this.count = 0;
        this.nestedMap.clear();
    }
    values() {
        return (Array.from(this.nestedMap.values()).flatMap((innerMap) => innerMap.values())[0] ?? oob());
    }
    [Symbol.iterator]() {
        return this.nestedMap[Symbol.iterator]();
    }
}
//# sourceMappingURL=nestedMap.js.map