/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
/**
 * RangeMap represents a mapping from integers to values of type T or undefined.
 * The values for a range of consecutive keys can be changed or queried in a single operation.
 */
export declare class RangeMap<T> {
    private readonly entries;
    constructor(initialEntries?: RangeEntry<T>[]);
    /**
     * Retrieves all entries from the rangeMap.
     * @returns An array of RangeEntryResult objects, each containing the start index, length, and value of a contiguous range.
     */
    getAllEntries(): readonly RangeQueryResult<T>[];
    /**
     * Retrieves the value for some prefix of the query range.
     *
     * @param start - The first key in the query range.
     * @param length - The length of the query range.
     * @returns A RangeQueryResult containing the value associated with `start`,
     * and the number of consecutive keys with that same value.
     */
    get(start: number, length: number): RangeQueryResult<T>;
    /**
     * Sets the value for a specified range.
     *
     * @param start - The first key in the range being set.
     * @param length - The length of the range.
     * @param value - The value to associate with the range.
     */
    set(start: number, length: number, value: T | undefined): void;
    /**
     * Deletes values within a specified range, updating or removing existing entries.
     *
     * 1. If an entry is completely included in the deletion range, the whole entry will be deleted
     * e.g.: map = [[1, 2], [4, 6]], delete range: [3, 6]
     * map becomes [[1, 2]] after deletion
     * (Note: the notation [a, b] represents start = a, end = b for simpler visualization, instead of `b`
     * representing the length)
     *
     * 2. If an entry is partially overlapped with the deletion range, the start or end point will be shifted
     * e.g.: map = [[1, 2], [4, 6]], delete range: [2, 4]
     * map becomes [[1, 1], [5, 6]] after deletion
     *
     * 3. If an entry completely includes the deletion range, the original entry may be split into two.
     * e.g.: map = [[1, 6]], delete range: [2, 4]
     * map becomes [[1, 1], [5, 6]]
     *
     * @param start - The start of the range to delete (inclusive).
     * @param length - The length of the range to delete.
     */
    delete(start: number, length: number): void;
}
/**
 * Represents a contiguous range of values in the RangeMap.
 * This interface is used internally and should not be exposed to consumers.
 */
interface RangeEntry<T> {
    /**
     * The starting index of the range (inclusive).
     */
    readonly start: number;
    /**
     * The length of the range.
     */
    readonly length: number;
    /**
     * The value associated with this range.
     */
    readonly value: T;
}
/**
 * Describes the result of a range query, including the value and length of the matching prefix.
 */
export interface RangeQueryResult<T> {
    /**
     * The value of the first key in the query range.
     * If no matching range is found, this will be undefined.
     */
    readonly value: T | undefined;
    /**
     * The length of the prefix of the query range which has the same value.
     * For example, if a RangeMap has the same value for keys 5, 6, and 7,
     * a query about the range [5, 10] would give a result with length 3.
     */
    readonly length: number;
}
export {};
//# sourceMappingURL=rangeMap.d.ts.map