/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */

import { assert } from "@fluidframework/core-utils/internal";
import type { IIdCompressor } from "@fluidframework/id-compressor";
import { assertIsStableId, isStableId } from "@fluidframework/id-compressor/internal";

import { brand, extractFromOpaque } from "../../util/index.js";

import type { LocalNodeKey, StableNodeKey } from "./nodeKey.js";

/**
 * An object which handles the generation of node keys as well as conversion between their two types ({@link StableNodeKey} and {@link LocalNodeKey}).
 */
export interface NodeKeyManager {
	/**
	 * Generate a {@link StableNodeKey}.
	 */
	generateLocalNodeKey(): LocalNodeKey;

	/**
	 * Convert the given {@link StableNodeKey} into its {@link LocalNodeKey} form.
	 */
	localizeNodeKey(key: StableNodeKey): LocalNodeKey;

	/**
	 * Convert the given {@link LocalNodeKey} into its {@link StableNodeKey} form.
	 */
	stabilizeNodeKey(key: LocalNodeKey): StableNodeKey;

	/**
	 * Attempts to recompress a StableNodeKey.
	 * @param key - The key that is attempted to recompress.
	 * @returns The `LocalNodeKey` associated with `key` or undefined if the key was not generated by any session known to this compressor.
	 */
	tryLocalizeNodeKey(key: string): LocalNodeKey | undefined;
}

/**
 * Creates a {@link NodeKeyManager} from the given {@link IIdCompressor}.
 * @param idCompressor - the compressor to use for key generation, compression, and decompression.
 * If undefined, then attempts to generate or convert keys will throw an error.
 */
export function createNodeKeyManager(
	idCompressor?: IIdCompressor | undefined,
): NodeKeyManager {
	return {
		generateLocalNodeKey: () => {
			assert(
				idCompressor !== undefined,
				0x6e4 /* Runtime IdCompressor must be available to generate local node keys */,
			);
			return brand(idCompressor.generateCompressedId());
		},

		localizeNodeKey: (key: StableNodeKey) => {
			assert(
				idCompressor !== undefined,
				0x6e5 /* Runtime IdCompressor must be available to convert node keys */,
			);
			return brand(idCompressor.recompress(key));
		},

		stabilizeNodeKey: (key: LocalNodeKey) => {
			assert(
				idCompressor !== undefined,
				0x6e6 /* Runtime IdCompressor must be available to convert node keys */,
			);
			return brand(
				// TODO: The assert below is required for type safety but is maybe slow
				assertIsStableId(idCompressor.decompress(extractFromOpaque(key))),
			);
		},
		tryLocalizeNodeKey: (key: string) => {
			assert(
				idCompressor !== undefined,
				0x6e9 /* Runtime IdCompressor must be available to convert node keys */,
			);
			if (isStableNodeKey(key)) {
				const compressedKey = idCompressor.tryRecompress(key);
				if (compressedKey !== undefined) {
					return brand(compressedKey);
				}
			}
		},
	};
}

export function isStableNodeKey(key: string): key is StableNodeKey {
	return isStableId(key);
}
